import pytest

from sunset import CallbackRegistry


class TestCallbackRegistry:
    def test_function(self):

        call_args: list[str] = []

        def test(value: str) -> None:
            call_args.append(value)

        reg: CallbackRegistry[str] = CallbackRegistry()

        assert len(reg) == 0
        reg.callAll("not added")
        assert call_args == []

        reg.add(test)
        assert len(reg) == 1
        assert test in reg
        reg.callAll("added")
        assert call_args == ["added"]

        del test
        assert len(reg) == 0
        reg.callAll("not added either")
        assert call_args == ["added"]

    def test_method(self):

        call_args: list[str] = []

        class Test:
            def test(self, value: str) -> None:
                call_args.append(value)

        reg: CallbackRegistry[str] = CallbackRegistry()

        t = Test()

        assert len(reg) == 0
        assert t.test not in reg

        reg.add(t.test)
        assert len(reg) == 1
        assert t.test in reg
        reg.callAll("added")
        assert call_args == ["added"]

        del t
        assert len(reg) == 0
        reg.callAll("not added")
        assert call_args == ["added"]

    def test_discard_function(self):
        def test(_: int) -> None:
            pass

        reg: CallbackRegistry[int] = CallbackRegistry()

        reg.add(test)
        assert test in reg
        reg.discard(test)
        assert test not in reg

    def test_discard_method(self):
        class Test:
            def test(self, _: int) -> None:
                pass

        t = Test()

        reg: CallbackRegistry[int] = CallbackRegistry()

        reg.add(t.test)
        assert t.test in reg
        reg.discard(t.test)
        assert t.test not in reg

    def test_pop(self):
        def test(_: bool):
            pass

        reg: CallbackRegistry[bool] = CallbackRegistry()

        reg.add(test)
        assert test in reg
        other = reg.pop()
        assert test not in reg
        assert other is test

    def test_function_added_once(self):

        call_args: list[str] = []

        def test(value: str) -> None:
            call_args.append(value)

        reg: CallbackRegistry[str] = CallbackRegistry()

        reg.add(test)
        assert len(reg) == 1
        reg.add(test)
        assert len(reg) == 1

        reg.callAll("once")
        assert call_args == ["once"]

    def test_same_method_added_once(self):

        call_args: list[str] = []

        class Test:
            def test(self, value: str) -> None:
                call_args.append(value)

        reg: CallbackRegistry[str] = CallbackRegistry()

        t = Test()

        reg.add(t.test)
        assert len(reg) == 1
        reg.add(t.test)
        assert len(reg) == 1

        reg.callAll("once")
        assert call_args == ["once"]

    def test_non_hashable_element(self):
        class Test(list[int]):
            def __init__(self) -> None:

                super().__init__()

                self.value = 0

            def test(self, value: int) -> None:
                self.value = value

        t = Test()

        with pytest.raises(TypeError):

            # Prove that type Test is not hashable.

            hash(t)

        reg: CallbackRegistry[int] = CallbackRegistry()

        reg.add(t.test)
        reg.callAll(42)

        assert t.value == 42

    def test_different_method_added_twice(self):

        call_args: list[str] = []

        class Test:
            def test(self, value: str) -> None:
                call_args.append(value)

        reg: CallbackRegistry[str] = CallbackRegistry()

        t1 = Test()
        t2 = Test()

        reg.add(t1.test)
        assert len(reg) == 1
        reg.add(t2.test)
        assert len(reg) == 2

        reg.callAll("twice")
        assert call_args == ["twice", "twice"]



import weakref

from typing import (
    Any,
    Callable,
    Generic,
    Iterable,
    Iterator,
    Optional,
    Type,
    TypeVar,
    cast,
)

from .exporter import maybe_escape
from .protocols import ContainableImpl
from .registry import CallbackRegistry
from .serializers import AnySerializableType, deserialize, serialize


# TODO: Replace with typing.Self when mypy finally supports that.
Self = TypeVar("Self", bound="Key[Any]")


class Key(Generic[AnySerializableType], ContainableImpl):
    """
    A single setting key containing a typed value.

    Keys support inheritance. If a Key does not have a value explicitly set, and
    it has a parent, then its value will be that of its parent. Else its value
    if unset is the default value it was instantiated with.

    Keys can call a callback when their value changes, regardless of if its
    their own value that changed, or that inherited from a parent. Set this
    callback with the :meth:`onValueChangeCall()` method.

    Args:
        default: (str, int, bool, float, or any type that implements the
            :class:`~sunset.Serializable` protocol) The value that this Key will
            return when not otherwise set; the type of this default determines
            the type of the values that can be set on this Key.

    Example:

    >>> from sunset import Key
    >>> key: Key[int] = Key(default=0)
    >>> key.get()
    0
    >>> key.set(42)
    >>> key.get()
    42
    >>> child_key: Key[int] = Key(default=0)
    >>> child_key.setParent(key)
    >>> child_key.get()
    42
    >>> child_key.set(101)
    >>> child_key.get()
    101
    >>> key.set(36)
    >>> key.get()
    36
    >>> child_key.get()
    101
    >>> child_key.clear()
    >>> child_key.get()
    36
    """

    _default: AnySerializableType
    _value: Optional[AnySerializableType]
    _value_change_callbacks: CallbackRegistry[AnySerializableType]
    _update_notification_callbacks: CallbackRegistry["Key[AnySerializableType]"]
    _update_notification_enabled: bool
    _parent: Optional[weakref.ref["Key[AnySerializableType]"]]
    _children: weakref.WeakSet["Key[AnySerializableType]"]
    _type: Type[AnySerializableType]

    def __init__(self, default: AnySerializableType):

        # serialize() raises an exception if the given value is not
        # serializable, so this call guarantees that the provided default is of
        # a type allowed in a Key. In case the user went ahead and ignored the
        # typechecker error when instantiating their Key with a bad default.

        serialize(default)

        super().__init__()

        self._default = default
        self._value = None

        self._value_change_callbacks = CallbackRegistry()
        self._update_notification_callbacks = CallbackRegistry()
        self._update_notification_enabled = True

        self._parent = None
        self._children = weakref.WeakSet()

        # Keep a runtime reference to the practical type contained in this
        # key.

        self._type = cast(Type[AnySerializableType], default.__class__)

    def get(self) -> AnySerializableType:
        """
        Returns the current value of this Key.

        If this Key does not currently have a value set on it, return that of
        its parent if any. If it does not have a parent, return the default
        value for this Key.

        Returns:
            This Key's current value.
        """

        if (value := self._value) is not None:
            return value

        if (parent := self.parent()) is not None:
            return parent.get()

        return self._default

    def set(self, value: AnySerializableType) -> None:
        """
        Sets the given value on this Key.

        Args:
            value: The value that this Key will now hold. Must be of the type
                bound to this Key, i.e. the same type as this Key's default
                value.

        Returns:
            None.
        """

        # Safety check in case the user is holding it wrong.

        if not isinstance(value, self._type):
            return

        previously_set = self.isSet()
        prev_value = self.get()

        self._value = value

        if prev_value != self.get():
            self._value_change_callbacks.callAll(self.get())
            for child in self.children():
                child._notifyParentValueChanged()

        if not previously_set or prev_value != self.get():
            self.triggerUpdateNotification()

    def clear(self) -> None:
        """
        Clears the value currently set on this Key, if any.

        Returns:
            None.
        """

        if not self.isSet():
            return

        prev_value = self.get()
        self._value = None

        if prev_value != self.get():
            self._value_change_callbacks.callAll(self.get())
            for child in self.children():
                child._notifyParentValueChanged()

        self.triggerUpdateNotification()

    def isSet(self) -> bool:
        """
        Returns whether there is a value currently set on this Key.

        Returns:
            True if a value is set on this Key, else False.
        """

        return self._value is not None

    def onValueChangeCall(
        self, callback: Callable[[AnySerializableType], None]
    ) -> None:
        """
        Adds a callback to be called whenever the value returned by calling
        :meth:`get()` on this Key would change, even if this Key itself was not
        updated; for instance, this will happen if there is no value currently
        set on it and its parent's value changed.

        The callback will be called with the new value as its argument.

        If you want a callback to be called whenever this Key is updated, even
        if its apparent value does not change, use :meth:`onUpdateCall()`
        instead. For example, if you call :meth:`set()` with a value of `0` on a
        Key newly created with a default value of `0`, callbacks added with
        :meth:`onUpdateCall()` are called and callbacks added with
        :meth:`onValueChangeCall()` are not.


        Args:
            callback: A callable that takes one argument of the same type of the
                values held by this Key, and that returns None.

        Returns:
            None.

        Note:
            This method does not increase the reference count of the given
            callback.
        """

        self._value_change_callbacks.add(callback)

    def onUpdateCall(
        self, callback: Callable[["Key[AnySerializableType]"], None]
    ) -> None:
        """
        Adds a callback to be called whenever this Key is updated, even if the
        value returned by :meth:`get()` does not end up changing.

        The callback will be called with this Key instance as its argument.

        If you want a callback to be called only when the apparent value of this
        Key changes, use :meth:`onValueChangeCall()` instead. For example, if a
        Key has no value set on it and has a parent whose value is updated, then
        callbacks added on this Key with :meth:`onValueChangeCall()` are called,
        and callbacks added with :meth:`onUpdateCall()` are not, because it's
        not *this* Key that was updated.

        Args:
            callback: A callable that will be called with one argument of type
                :class:`~sunset.Key`.

        Returns:
            None.

        Note:
            This method does not increase the reference count of the given
            callback.
        """

        self._update_notification_callbacks.add(callback)

    def setParent(self: Self, parent: Optional[Self]) -> None:
        """
        Makes the given Key the parent of this one. If None, remove this
        Key's parent, if any.

        A Key with a parent will inherit its parent's value when this
        Key's own value is not currently set.

        This method is for internal purposes and you will typically not need to
        call it directly.

        Args:
            parent: Either a Key that will become this Key's parent, or
                None. The parent Key must have the same type as this
                Key.

        Returns:
            None.

        Note:
            A Key and its parent, if any, do not increase each other's
            reference count.
        """

        # Runtime check to affirm the type check of the method.

        if parent is not None:
            if type(self) is not type(parent):
                return

        old_parent = self.parent()
        if old_parent is not None:
            old_parent._children.discard(self)

        if parent is None:
            self._parent = None
            return

        if parent._type is not self._type:

            # This should not happen... unless the user is holding it wrong.
            # So, better safe than sorry.

            return

        parent._children.add(self)
        self._parent = weakref.ref(parent)

    def parent(self: Self) -> Optional[Self]:
        """
        Returns the parent of this Key, if any.

        Returns:
            A Key instance of the same type as this one, or None.
        """

        # Make the type of self._parent more specific for the purpose of type
        # checking.

        _parent = cast(Optional[weakref.ref[Self]], self._parent)
        return _parent() if _parent is not None else None

    def children(self: Self) -> Iterator[Self]:
        """
        Returns an iterator over the Keys that have this Key as their parent.

        Returns:
            An iterator over Keys of the same type as this one.
        """

        for child in self._children:
            yield cast(Self, child)

    def dumpFields(self) -> Iterable[tuple[str, Optional[str]]]:

        if self.isSet() and not self.isPrivate():
            yield self.fieldPath(), serialize(self.get())

    def restoreField(self, path: str, value: Optional[str]) -> None:

        if value is None:
            return

        self._update_notification_enabled = False

        if path == self.fieldLabel():
            if (val := deserialize(self._type, value)) is not None:
                self.set(val)

        self._update_notification_enabled = True

    def _notifyParentValueChanged(self):

        if self.isSet():
            return

        self._value_change_callbacks.callAll(self.get())
        for child in self.children():
            child._notifyParentValueChanged()

    def triggerUpdateNotification(self):
        """
        Internal.
        """

        if not self._update_notification_enabled:
            return

        self._update_notification_callbacks.callAll(self)

        if (container := self.container()) is not None and not self.isPrivate():
            container.triggerUpdateNotification(self)

    def newInstance(self: Self) -> Self:
        """
        Internal. Returns a new instance of this Key with the same default
        value.

        Returns:
            A new Key.
        """

        return self.__class__(default=self._default)

    def __repr__(self) -> str:

        type_name = self._type.__name__
        value = maybe_escape(serialize(self.get()))
        if not self.isSet():
            value = f"({value})"
        return f"<Key[{type_name}]:{value}>"



import dataclasses
import inspect
import weakref

from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    MutableSet,
    Optional,
    Type,
    TypeVar,
    cast,
)

from .non_hashable_set import WeakNonHashableSet
from .protocols import (
    Containable,
    ContainableImpl,
    Field,
    ItemTemplate,
    UpdateNotifier,
)
from .registry import CallbackRegistry


# TODO: Replace with typing.Self when mypy finally supports that.
Self = TypeVar("Self", bound="Bundle")


class Bundle(ContainableImpl):
    """
    A collection of related Keys.

    Under the hood, a Bundle is a dataclass, and can be used in the same manner,
    i.e. by defining attributes directly on the class itself.

    Example:

    >>> from sunset import Bundle, Key
    >>> class Appearance(Bundle):
    ...     class Font(Bundle):
    ...         name: Key[str] = Key(default="Arial")
    ...         size: Key[int] = Key(default=14)
    ...     main_font: Font      = Font()
    ...     secondary_font: Font = Font()
    >>> appearance = Appearance()
    >>> appearance.main_font.name.get()
    'Arial'
    >>> appearance.secondary_font.name.get()
    'Arial'
    >>> appearance.main_font.name.set("Times New Roman")
    >>> appearance.secondary_font.name.set("Calibri")
    >>> appearance.main_font.name.get()
    'Times New Roman'
    >>> appearance.secondary_font.name.get()
    'Calibri'
    """

    _parent: Optional[weakref.ref["Bundle"]]
    _children: MutableSet["Bundle"]
    _fields: dict[str, Field]
    _update_notification_callbacks: CallbackRegistry[UpdateNotifier]
    _update_notification_enabled: bool

    def __new__(cls: Type[Self]) -> Self:

        # Automatically promote relevant attributes to dataclass fields.

        cls_parents = cls.__bases__

        potential_fields = list(vars(cls).items())
        for name, attr in potential_fields:

            if inspect.isclass(attr):

                if attr.__name__ == name:

                    # This is probably a class definition that just happens to
                    # be located inside the containing Bundle definition. This
                    # is fine.

                    continue

                raise TypeError(
                    f"Field '{name}' in the definition of '{cls.__name__}'"
                    " is uninstantiated"
                )

            if isinstance(attr, ItemTemplate):

                # Safety check: make sure the user isn't accidentally overriding
                # an existing attribute.

                for cls_parent in cls_parents:
                    if getattr(cls_parent, name, None) is not None:
                        raise TypeError(
                            f"Field '{name}' in the definition of"
                            f" '{cls.__name__}' overrides attribute of the same"
                            f" name declared in parent class"
                            f" '{cls_parent.__name__}'; consider"
                            f" renaming this field to '{name}_' for instance"
                        )

                setattr(
                    cls,
                    name,
                    dataclasses.field(default_factory=attr.newInstance),
                )

                # Dataclass instantiation raises an error if a field does not
                # have an explicit type annotation. But our Key, List and
                # Bundle fields are unambiguously typed, so we don't actually
                # need the annotation. So we just tell the dataclass that the
                # type of non-explicitly-annotated fields is 'Any'. Turns out,
                # this works.

                # Also note the subtle dance here: the annotations need to be on
                # *this* class, and not inherited from a parent class. So we
                # make sure that the __annotations__ mapping does exist in this
                # class' namespace.

                if "__annotations__" not in cls.__dict__:
                    setattr(cls, "__annotations__", {})
                cls.__annotations__.setdefault(name, Any)

        # Create a new instance of this class wrapped as a dataclass.

        wrapped = dataclasses.dataclass()(cls)
        return super().__new__(wrapped)

    def __post_init__(self) -> None:

        super().__init__()

        self._parent = None
        self._children = WeakNonHashableSet[Bundle]()
        self._fields = {}
        self._update_notification_callbacks = CallbackRegistry()
        self._update_notification_enabled = True

        for label, field in vars(self).items():

            if isinstance(field, Field):
                self._fields[label] = field
                field.setContainer(label, self)

    def fieldPath(self) -> str:
        """
        Internal.
        """

        return super().fieldPath() + self._PATH_SEPARATOR

    def containsFieldWithLabel(self, label: str, field: Containable) -> bool:
        """
        Internal.
        """

        return self._fields.get(label) is field

    def setParent(self: Self, parent: Optional[Self]) -> None:
        """
        Makes the given Bundle the parent of this one. If None, remove this
        Bundle's parent, if any.

        All the Key, List and Bundle fields defined on this Bundle instance
        will be recursively reparented to the corresponding Key / List / Bundle
        field on the given parent.

        This method is for internal purposes and you will typically not need to
        call it directly.

        Args:
            parent: Either a Bundle that will become this Bundle's parent, or
                None. The parent Bundle must have the same type as this
                Bundle.

        Returns:
            None.

        Note:
            A Bundle and its parent, if any, do not increase each other's
            reference count.
        """

        # Runtime check to affirm the type check of the method.

        if parent is not None:
            if type(self) is not type(parent):
                return

        old_parent = self.parent()
        if old_parent is not None:
            old_parent._children.discard(self)

        if parent is None:
            self._parent = None
        else:
            self._parent = weakref.ref(parent)
            parent._children.add(self)

        for label, field in self._fields.items():

            if parent is None:
                field.setParent(None)
                continue

            parent_field = parent._fields.get(label)
            if parent_field is None:

                # This is a safety check, but it shouldn't happen. By
                # construction self should be of the same type as parent, so
                # they should have the same attributes.

                continue

            assert type(field) is type(parent_field)
            field.setParent(parent_field)

    def parent(self: Self) -> Optional[Self]:
        """
        Returns the parent of this Bundle, if any.

        Returns:
            A Bundle instance of the same type as this one, or None.
        """

        # Make the type of self._parent more specific for the purpose of type
        # checking.

        _parent = cast(Optional[weakref.ref[Self]], self._parent)
        return _parent() if _parent is not None else None

    def children(self: Self) -> Iterator[Self]:
        """
        Returns an iterator over the Bundle instances that have this Bundle
        as their parent.

        Returns:
            An iterator over Bundle instances of the same type as this one.
        """

        # Note that we iterate on a copy so that this will not break if a
        # different thread updates the contents during the iteration.

        for child in list(self._children):
            yield cast(Self, child)

    def onUpdateCall(self, callback: Callable[[Any], None]) -> None:
        """
        Adds a callback to be called whenever this Bundle is updated. A Bundle
        is considered updated when any of its fields is updated.

        The callback will be called with as its argument whichever field was
        just updated.

        Args:
            callback: A callable that will be called with one argument of type
                :class:`~sunset.List`, :class:`~sunset.Bundle` or
                :class:`~sunset.Key`, and returns None.

        Returns:
            None.

        Note:
            This method does not increase the reference count of the given
            callback.
        """

        self._update_notification_callbacks.add(callback)

    def isSet(self) -> bool:
        """
        Indicates whether this Bundle holds any field that is set.

        Returns:
            True if any field set on this Bundle is set, else False.
        """

        return any(field.isSet() for field in self._fields.values())

    def dumpFields(self) -> Iterable[tuple[str, Optional[str]]]:
        """
        Internal.
        """

        if not self.isPrivate():
            for _, field in sorted(self._fields.items()):
                yield from field.dumpFields()

    def restoreField(self, path: str, value: Optional[str]) -> None:
        """
        Internal.
        """

        if self._PATH_SEPARATOR not in path:
            return

        field_label, path = path.split(self._PATH_SEPARATOR, 1)
        if self.fieldLabel() != field_label:
            return

        field_label, *_ = path.split(self._PATH_SEPARATOR, 1)
        if (field := self._fields.get(field_label)) is not None:
            field.restoreField(path, value)

    def triggerUpdateNotification(
        self, field: Optional[UpdateNotifier]
    ) -> None:
        """
        Internal.
        """

        if not self._update_notification_enabled:
            return

        if field is None:
            field = self

        self._update_notification_callbacks.callAll(field)

        if (container := self.container()) is not None and not self.isPrivate():
            container.triggerUpdateNotification(field)

    def newInstance(self: Self) -> Self:
        """
        Internal. Returns a new instance of this Bundle with the same fields.

        Returns:
            A Bundle instance.
        """

        new = self.__class__()
        return new



from typing import Optional

from sunset import serializers


class ExampleSerializable:
    def __init__(self, value: str) -> None:

        self._value = value

    def toStr(self) -> str:

        return self._value

    @staticmethod
    def fromStr(value: str) -> Optional["ExampleSerializable"]:

        return ExampleSerializable(value)


def test_serialize_int():

    assert serializers.serialize(42) == "42"


def test_serialize_str():

    assert serializers.serialize(" !! test !!") == " !! test !!"


def test_serialize_bool():

    assert serializers.serialize(True) == "true"
    assert serializers.serialize(False) == "false"


def test_serialize_float():

    assert serializers.serialize(1.0) == "1.0"
    assert serializers.serialize(2.34e-56) == "2.34e-56"


def test_serialize_serializable():

    t = ExampleSerializable("serializable")
    assert serializers.serialize(t) == "serializable"


def test_deserialize_str():

    assert serializers.deserialize(str, "test") == "test"


def test_deserialize_int():

    assert serializers.deserialize(int, "test") is None
    assert serializers.deserialize(int, "   -32 ") == -32
    assert serializers.deserialize(int, "00017") == 17


def test_deserialize_float():

    assert serializers.deserialize(float, "test") is None
    assert serializers.deserialize(float, "   -32.10 ") == -32.1
    assert serializers.deserialize(float, "2.34e-56") == 2.34e-56


def test_deserialize_bool():

    assert serializers.deserialize(bool, "YES")
    assert serializers.deserialize(bool, "  1 ")
    assert serializers.deserialize(bool, "true")

    assert not serializers.deserialize(bool, "NO")
    assert not serializers.deserialize(bool, "  0 ")
    assert not serializers.deserialize(bool, "false")

    assert serializers.deserialize(bool, "garbage") is None


def test_deserialize_serializable():

    t = serializers.deserialize(ExampleSerializable, "this is a test")
    assert t is not None
    assert isinstance(t, ExampleSerializable)
    assert t.toStr() == "this is a test"



import weakref

from collections.abc import MutableSet
from types import MethodType
from typing import Any, Callable, Iterator, TypeVar

from .non_hashable_set import NonHashableSet

_T = TypeVar("_T")


class CallbackRegistry(MutableSet[Callable[[_T], None]]):

    _content: NonHashableSet[weakref.ref[Callable[[_T], None]]]

    def __init__(self) -> None:

        super().__init__()

        # We can't just use a WeakNonHashableSet because it does not know how to
        # take references to methods. Instead we use a regular NonHashableSet
        # and manually take the proper type of weak reference when adding a
        # callable to the set.

        self._content = NonHashableSet()

    def add(self, value: Callable[[_T], None]) -> None:

        if isinstance(value, MethodType):

            # Note: WeakMethod has incorrect type annotations, so we have to
            # ignore types here.

            r = weakref.WeakMethod(value, self._onExpire)  # type: ignore

        else:
            r = weakref.ref(value, self._onExpire)

        self._content.add(r)

    def __contains__(self, value: Any) -> bool:

        return any(self._isSameCallable(candidate, value) for candidate in self)

    def __iter__(self) -> Iterator[Callable[[_T], None]]:

        for ref in self._content:
            value = ref()
            if value is not None:
                yield value

    def __len__(self) -> int:

        return len(self._content)

    def discard(self, value: Callable[[_T], None]) -> None:

        refs = list(self._content)
        for ref in refs:
            callable = ref()
            if callable is not None and self._isSameCallable(callable, value):
                self._content.discard(ref)

    def callAll(self, value: _T) -> None:

        for callback in self:
            callback(value)

    @staticmethod
    def _isSameCallable(
        callable1: Callable[[_T], None], callable2: Callable[[_T], None]
    ) -> bool:

        if isinstance(callable1, MethodType) and isinstance(
            callable2, MethodType
        ):
            return (
                callable1.__self__ is callable2.__self__
                and callable1.__name__ == callable2.__name__
            )

        else:
            return callable1 is callable2

    def _onExpire(self, ref: weakref.ref[Callable[[_T], None]]) -> None:

        self._content.discard(ref)



"SunsetSettings: a type-safe, extensible settings system with inheritance."

__project__ = "SunsetSettings"
__version__ = "0.3.1"
__author__ = "P. Varet"
__copyright__ = "2022, P. Varet"

from . import exporter, non_hashable_set, serializers

from .bundle import Bundle
from .key import Key
from .list import List
from .protocols import Serializable
from .registry import CallbackRegistry
from .settings import Settings, normalize

__all__ = [
    "Bundle",
    "CallbackRegistry",
    "Key",
    "List",
    "Serializable",
    "Settings",
    "exporter",
    "non_hashable_set",
    "normalize",
    "serializers",
]



from typing import Any, Optional, Type, TypeVar, Union, cast

from .protocols import Serializable

_AnySerializableType = Union[int, str, bool, float, Serializable]

AnySerializableType = TypeVar("AnySerializableType", bound=_AnySerializableType)


def serialize(value: _AnySerializableType) -> str:

    if isinstance(value, bool):
        return "true" if value else "false"
    elif isinstance(value, int) or isinstance(value, float):
        return str(value)
    elif isinstance(value, str):
        return value
    elif isinstance(cast(Any, value), Serializable):

        # Note the cast above. It's just so that linters don't complain that
        # we're using an isinstance() check when that's the only type the value
        # can possibly be, from the typechecker's point of view. But the
        # TypeError below is explicitly about catching this kind of user error.
        # So we do in fact want this isinstance() check.

        return value.toStr()

    else:
        raise TypeError(
            f"'{repr(value)}' is of type '{value.__class__.__name__}', which is"
            " not serializable"
        )


def deserialize(
    _type: Type[AnySerializableType], string: str
) -> Optional[AnySerializableType]:

    if issubclass(_type, bool):
        if string.strip().lower() in ("true", "yes", "y", "1"):
            # The cast is unnecessary, but works around a mypy bug.
            return cast(AnySerializableType, _type(True))
        if string.strip().lower() in ("false", "no", "n", "0"):
            # The cast is unnecessary, but works around a mypy bug.
            return cast(AnySerializableType, _type(False))
        return None

    if issubclass(_type, int):
        try:
            # The cast is unnecessary, but works around a mypy bug.
            return cast(AnySerializableType, _type(int(string)))
        except ValueError:
            return None

    if issubclass(_type, float):
        try:
            # The cast is unnecessary, but works around a mypy bug.
            return cast(AnySerializableType, _type(float(string)))
        except ValueError:
            return None

    if issubclass(_type, str):
        # The cast is unnecessary, but works around a mypy bug.
        return cast(AnySerializableType, _type(string))

    assert issubclass(_type, Serializable)

    # The cast is unnecessary, but works around a mypy bug.
    return cast(AnySerializableType, _type.fromStr(string))



import pytest

from pytest_mock import MockerFixture

from sunset import Bundle, Key, List, protocols


class ExampleBundle(Bundle):
    class InnerBundle(Bundle):
        b = Key(42)

    class Item(Bundle):
        c = Key("default c")

    a = Key("default a")
    inner_bundle = InnerBundle()
    list = List(Item(), order=List.PARENT_LAST)
    _private = Key("private")


class TestBundle:
    def test_protocol_implementation(self):

        bundle = ExampleBundle()
        assert isinstance(bundle, protocols.Field)
        assert isinstance(bundle, protocols.Container)

    def test_creation(self):

        bundle = ExampleBundle()
        bundle.list.appendOne()
        assert bundle.a.get() == "default a"
        assert bundle.inner_bundle.b.get() == 42
        assert bundle.list[0].c.get() == "default c"

    def test_uninstantiated_field_fails(self):
        class InnerBundle(Bundle):
            pass

        class FaultyBundle(Bundle):
            inner = InnerBundle

        with pytest.raises(TypeError):
            FaultyBundle()

    def test_inner_bundle_definition_is_fine(self):
        class FineBundle(Bundle):
            class InnerBundle(Bundle):
                pass

            inner = InnerBundle()

        FineBundle()

    def test_fields_cant_override_existing_attributes(self):

        # "__init__" is an attribute that happens to exist on the class.

        assert getattr(Bundle, "__init__", None) is not None

        class FaultyBundle(Bundle):
            __init__ = Key(default="test")  # type: ignore[assignment]

        with pytest.raises(TypeError):
            FaultyBundle()

    def test_inheritance(self):

        parent_bundle = ExampleBundle()
        child_bundle = ExampleBundle()

        assert child_bundle not in parent_bundle.children()
        assert child_bundle.parent() is None

        child_bundle.setParent(parent_bundle)
        assert child_bundle in parent_bundle.children()
        assert child_bundle.parent() is parent_bundle

        child_bundle.setParent(None)
        assert child_bundle not in parent_bundle.children()
        assert child_bundle.parent() is None

    def test_parenting(self):

        parent_bundle = ExampleBundle()
        child_bundle = ExampleBundle()
        child_bundle.setParent(parent_bundle)

        assert child_bundle.parent() is parent_bundle
        assert child_bundle in parent_bundle.children()

    def test_reparenting(self):

        bundle1 = ExampleBundle()
        bundle2 = ExampleBundle()
        child_bundle = ExampleBundle()

        assert child_bundle not in bundle1.children()
        assert child_bundle not in bundle2.children()
        assert child_bundle.parent() is None

        child_bundle.setParent(bundle1)
        assert child_bundle in bundle1.children()
        assert child_bundle not in bundle2.children()
        assert child_bundle.parent() is bundle1

        child_bundle.setParent(bundle2)
        assert child_bundle not in bundle1.children()
        assert child_bundle in bundle2.children()
        assert child_bundle.parent() is bundle2

        child_bundle.setParent(None)
        assert child_bundle not in bundle1.children()
        assert child_bundle not in bundle2.children()
        assert child_bundle.parent() is None

    def test_inheritance_propagation(self):

        parent_bundle = ExampleBundle()
        child_bundle = ExampleBundle()
        child_bundle.setParent(parent_bundle)

        assert child_bundle.a.parent() is parent_bundle.a
        assert child_bundle.inner_bundle.parent() is parent_bundle.inner_bundle
        assert (
            child_bundle.inner_bundle.b.parent() is parent_bundle.inner_bundle.b
        )
        assert child_bundle.list.parent() is parent_bundle.list

    def test_inheritance_values(self):

        parent_bundle = ExampleBundle()
        child_bundle = ExampleBundle()
        child_bundle.setParent(parent_bundle)

        parent_bundle.a.set("test parent")
        assert child_bundle.a.get() == "test parent"
        child_bundle.a.set("test child")
        assert child_bundle.a.get() == "test child"
        assert parent_bundle.a.get() == "test parent"

        parent_bundle.inner_bundle.b.set(101)
        assert child_bundle.inner_bundle.b.get() == 101
        child_bundle.inner_bundle.b.set(37)
        assert child_bundle.inner_bundle.b.get() == 37
        assert parent_bundle.inner_bundle.b.get() == 101

        parent_bundle.list.appendOne().c.set("test parent")
        assert [item.c.get() for item in parent_bundle.list.iter()] == [
            "test parent"
        ]
        assert [item.c.get() for item in child_bundle.list.iter()] == [
            "test parent"
        ]
        child_bundle.list.appendOne().c.set("test child")
        assert [item.c.get() for item in parent_bundle.list.iter()] == [
            "test parent"
        ]
        assert [item.c.get() for item in child_bundle.list.iter()] == [
            "test child",
            "test parent",
        ]
        del parent_bundle.list[0]
        assert [item.c.get() for item in parent_bundle.list.iter()] == []
        assert [item.c.get() for item in child_bundle.list.iter()] == [
            "test child"
        ]

    def test_field_label(self):

        bundle = ExampleBundle()

        assert bundle.fieldLabel() == ""
        assert bundle.inner_bundle.fieldLabel() == "inner_bundle"

    def test_field_path(self):

        bundle = ExampleBundle()
        assert bundle.fieldPath() == "."
        assert bundle.a.fieldPath() == ".a"
        assert bundle.inner_bundle.fieldPath() == ".inner_bundle."
        assert bundle.inner_bundle.b.fieldPath() == ".inner_bundle.b"
        assert bundle.inner_bundle.b.fieldPath() == ".inner_bundle.b"
        assert bundle.list.fieldPath() == ".list."
        bundle.list.appendOne()
        assert bundle.list[0].fieldPath() == ".list.1."
        assert bundle.list[0].c.fieldPath() == ".list.1.c"

    def test_dump_fields(self):

        bundle = ExampleBundle()
        assert list(bundle.dumpFields()) == []

        bundle.a.set("test dump a")
        bundle.inner_bundle.b.set(101)
        bundle.list.appendOne().c.set("test dump c 1")
        bundle.list.appendOne()
        bundle.list.appendOne().c.set("test dump c 3")
        bundle._private.set("test private")  # type: ignore

        assert list(bundle.dumpFields()) == [
            (".a", "test dump a"),
            (".inner_bundle.b", "101"),
            (".list.1.c", "test dump c 1"),
            (".list.2", None),
            (".list.3.c", "test dump c 3"),
        ]

    def test_restore_field(self, mocker: MockerFixture):

        callback = mocker.stub()

        # Test all flavors of valid paths. Also, restoring a field should not
        # trigger a callbacak.

        bundle = ExampleBundle()
        bundle.onUpdateCall(callback)

        assert bundle.a.get() == "default a"
        bundle.restoreField(".a", "restore a")
        assert bundle.a.get() == "restore a"
        callback.assert_not_called()

        assert len(bundle.list) == 0
        bundle.restoreField(".list.2.c", "restore c 2")
        bundle.restoreField(".list.1.c", "restore c 1")
        assert len(bundle.list) == 2
        assert bundle.list[0].c.get() == "restore c 1"
        assert bundle.list[1].c.get() == "restore c 2"
        callback.assert_not_called()

        bundle.restoreField(".list.1.c", "other restore c 1")
        assert len(bundle.list) == 2
        assert bundle.list[0].c.get() == "other restore c 1"
        assert bundle.list[1].c.get() == "restore c 2"
        callback.assert_not_called()

        assert bundle.inner_bundle.b.get() == 42
        bundle.restoreField(".inner_bundle.b", "101")
        assert bundle.inner_bundle.b.get() == 101
        callback.assert_not_called()

        # Test invalid paths.

        other_bundle = ExampleBundle()
        other_bundle.onUpdateCall(callback)

        other_bundle.restoreField("a", "invalid path")
        assert other_bundle.a.get() == "default a"
        assert not other_bundle.isSet()

        other_bundle.restoreField("a.", "invalid path")
        assert other_bundle.a.get() == "default a"
        assert not other_bundle.isSet()

        other_bundle.restoreField(".", "invalid path")
        assert not other_bundle.isSet()

        other_bundle.restoreField(".invalid", "invalid path")
        assert not other_bundle.isSet()

        other_bundle.restoreField("", "invalid path")
        assert not other_bundle.isSet()

    def test_persistence(self):

        # A Bundle does not keep a reference to its parent or children.

        bundle = ExampleBundle()
        level1 = ExampleBundle()
        level2 = ExampleBundle()
        level1.setParent(bundle)
        level2.setParent(level1)

        assert level1.parent() is not None
        assert len(list(level1.children())) == 1

        del bundle
        del level2
        assert level1.parent() is None
        assert len(list(level1.children())) == 0

    def test_key_updated_notification(self, mocker: MockerFixture):

        callback = mocker.stub()

        bundle = ExampleBundle()
        bundle.onUpdateCall(callback)

        child = ExampleBundle()
        child.setParent(bundle)

        bundle.a.set("test 1")
        callback.assert_called_once_with(bundle.a)
        callback.reset_mock()

        bundle.inner_bundle.b.set(123)
        callback.assert_called_once_with(bundle.inner_bundle.b)
        callback.reset_mock()

        bundle.list.appendOne()
        callback.assert_called_once_with(bundle.list)
        callback.reset_mock()

        bundle.list[0].c.set("test 2")
        callback.assert_called_once_with(bundle.list[0].c)
        callback.reset_mock()

        child.a.set("test 3")
        callback.assert_not_called()
        callback.reset_mock()



import weakref

from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    Optional,
    Protocol,
    Type,
    TypeVar,
    runtime_checkable,
)


Self = TypeVar("Self")


@runtime_checkable
class Serializable(Protocol):
    """
    A protocol to be implemented by a class in order to enable storing instances
    of that class in a Key.

    The two methods to be implemented are :meth:`toStr()` and (classmethod)
    :meth:`fromStr()`.
    """

    def toStr(self) -> str:
        """
        Returns a string representation of this instance that can be used by
        :meth:`fromStr()` to reconstruct a copy of this instance.
        """
        ...

    @classmethod
    def fromStr(cls: Type[Self], value: str) -> Optional[Self]:
        """
        Takes a string that represents a serialized instance of this class, and
        returns a newly created instance that corresponds to that
        representation, or None if the string is not a valid serialized
        representation of an instance of this class.
        """
        ...


@runtime_checkable
class Inheriter(Protocol):
    def setParent(self: Self, parent: Optional[Self]) -> None:
        ...

    def parent(self: Self) -> Optional[Self]:
        ...

    def children(self: Self) -> Iterator[Self]:
        ...


@runtime_checkable
class Dumpable(Protocol):
    def dumpFields(self) -> Iterable[tuple[str, Optional[str]]]:
        ...

    def restoreField(self, path: str, value: Optional[str]) -> None:
        ...

    def isSet(self) -> bool:
        ...


@runtime_checkable
class UpdateNotifier(Protocol):
    def onUpdateCall(self, callback: Callable[[Any], None]) -> None:
        ...


@runtime_checkable
class ItemTemplate(Protocol):
    def newInstance(self: Self) -> Self:
        ...


@runtime_checkable
class Containable(Protocol):

    _PATH_SEPARATOR: str

    def setContainer(
        self, label: str, container: Optional["Container"]
    ) -> None:
        ...

    def container(self) -> Optional["Container"]:
        ...

    def fieldLabel(self) -> str:
        ...

    def fieldPath(self) -> str:
        ...

    def isPrivate(self) -> bool:
        ...


@runtime_checkable
class Container(Containable, UpdateNotifier, Protocol):
    def containsFieldWithLabel(self, label: str, field: "Containable") -> bool:
        ...

    def triggerUpdateNotification(
        self, field: "Optional[UpdateNotifier]"
    ) -> None:
        ...


class ContainableImpl:
    """
    A ready-to-use implementation of the Containable protocol.
    """

    _PATH_SEPARATOR: str = "."

    _field_label: str
    _container: Optional[weakref.ref[Container]]

    def __init__(self) -> None:

        super().__init__()

        self._field_label = ""
        self._container = None

    def setContainer(self, label: str, container: Optional[Container]) -> None:
        """
        Internal.
        """

        if container is None:
            self._container = None
            self._field_label = ""
        else:
            self._container = weakref.ref(container)
            self._field_label = label

    def container(self) -> Optional[Container]:
        """
        Internal.
        """

        if self._container is None:
            return None

        container = self._container()
        if container is None:
            return None

        # Make sure this Containable is in fact still held in its supposed
        # Container. Else update the situation.

        if not container.containsFieldWithLabel(self._field_label, self):
            self.setContainer("", None)
            return None

        return container

    def fieldLabel(self) -> str:
        """
        Internal.
        """

        if self.container() is None:
            self._field_label = ""

        return self._field_label

    def fieldPath(self) -> str:
        """
        Internal.
        """

        path = (
            ""
            if (container := self.container()) is None
            else container.fieldPath()
        )

        return path + self.fieldLabel()

    def isPrivate(self) -> bool:
        """
        Internal.
        """

        return self.fieldLabel().startswith("_")


assert isinstance(ContainableImpl, Containable)


@runtime_checkable
class Field(
    Containable,
    Dumpable,
    Inheriter,
    ItemTemplate,
    UpdateNotifier,
    Protocol,
):
    ...



from typing import Iterator

from pytest_mock import MockerFixture

from sunset import Bundle, Key, List, protocols


class ExampleBundle(Bundle):

    test = Key("")


class TestList:
    def test_protocol_implementation(self):

        list_key: List[Key[int]] = List(Key(default=0))
        assert isinstance(list_key, protocols.Field)
        assert isinstance(list_key, protocols.Container)

    def test_add_pop(self):

        list_key: List[Key[str]] = List(Key(default=""))
        list_key.appendOne()

        assert len(list_key) == 1

        list_key[0].set("test")
        assert list_key[0].get() == "test"

        list_key.pop()

        assert len(list_key) == 0

    def test_inheritance(self):

        parent_list: List[Key[int]] = List(Key(default=0))
        child_list: List[Key[int]] = List(Key(default=0))

        assert child_list not in parent_list.children()
        assert child_list.parent() is None

        child_list.setParent(parent_list)
        assert child_list in parent_list.children()
        assert child_list.parent() is parent_list

        child_list.setParent(None)
        assert child_list not in parent_list.children()
        assert child_list.parent() is None

    def test_reparenting(self):

        list1: List[Key[int]] = List(Key(default=0))
        list2: List[Key[int]] = List(Key(default=0))
        child_list: List[Key[int]] = List(Key(default=0))

        assert child_list not in list1.children()
        assert child_list not in list2.children()
        assert child_list.parent() is None

        child_list.setParent(list1)
        assert child_list in list1.children()
        assert child_list not in list2.children()
        assert child_list.parent() is list1

        child_list.setParent(list2)
        assert child_list not in list1.children()
        assert child_list in list2.children()
        assert child_list.parent() is list2

        child_list.setParent(None)
        assert child_list not in list1.children()
        assert child_list not in list2.children()
        assert child_list.parent() is None

    def test_iter_inheritance(self):
        def flatten(keys: Iterator[Key[str]]) -> list[str]:
            return [key.get() for key in keys]

        parent: List[Key[str]] = List(Key(default=""))
        child_default: List[Key[str]] = List(Key(default=""))
        child_iter_no_parent: List[Key[str]] = List(
            Key(default=""), order=List.NO_PARENT
        )
        child_iter_parent_first: List[Key[str]] = List(
            Key(default=""), order=List.PARENT_FIRST
        )
        child_iter_parent_last: List[Key[str]] = List(
            Key(default=""), order=List.PARENT_LAST
        )

        parent.appendOne().set("parent")
        child_default.appendOne().set("child")
        child_iter_no_parent.appendOne().set("child")
        child_iter_parent_first.appendOne().set("child")
        child_iter_parent_last.appendOne().set("child")

        assert flatten(child_default.iter()) == ["child"]

        child_default.setParent(parent)
        child_iter_no_parent.setParent(parent)
        child_iter_parent_first.setParent(parent)
        child_iter_parent_last.setParent(parent)

        assert flatten(child_default.iter()) == ["child"]
        assert flatten(child_iter_no_parent.iter()) == ["child"]
        assert flatten(child_iter_parent_first.iter()) == ["parent", "child"]
        assert flatten(child_iter_parent_last.iter()) == ["child", "parent"]
        assert flatten(child_default.iter(order=List.NO_PARENT)) == ["child"]
        assert flatten(child_default.iter(order=List.PARENT_FIRST)) == [
            "parent",
            "child",
        ]
        assert flatten(child_default.iter(order=List.PARENT_LAST)) == [
            "child",
            "parent",
        ]

    def test_dump_fields(self):

        key_list = List(Key(""))
        key_list.appendOne().set("test 1")
        key_list.appendOne()
        key_list.appendOne().set("test 3")
        assert list(key_list.dumpFields()) == [
            (".1", "test 1"),
            (".2", None),
            (".3", "test 3"),
        ]

        bundle_list = List(ExampleBundle())
        bundle_list.appendOne().test.set("test 1")
        bundle_list.appendOne()
        bundle_list.appendOne().test.set("test 3")
        bundle_list.appendOne()
        assert list(bundle_list.dumpFields()) == [
            (".1.test", "test 1"),
            (".2", None),
            (".3.test", "test 3"),
            (".4", None),
        ]

        key_list = List(Key(""))
        key_list.appendOne()
        key_list.appendOne().set("")
        key_list.appendOne()
        assert list(key_list.dumpFields()) == [
            (".1", None),
            (".2", ""),
            (".3", None),
        ]

        class TestBundle(Bundle):

            key_list = List(Key(""))
            _private = List(Key(""))

        bundle = TestBundle()
        bundle.key_list.appendOne().set("test public")
        bundle._private.appendOne().set("test private")  # type: ignore
        assert list(bundle.dumpFields()) == [
            (".key_list.1", "test public"),
        ]

    def test_restore_field(self, mocker: MockerFixture):

        callback = mocker.stub()

        # Test restoring one value. Also, restoring a field should not trigger a
        # callback.

        test_list1 = List(Key("default"))
        test_list1.onUpdateCall(callback)
        test_list1.restoreField(".1", "test")
        assert len(test_list1) == 1
        assert test_list1[0].get() == "test"
        callback.assert_not_called()

        # Test overwriting a restored value.

        test_list1.restoreField(".1", "other test")
        assert len(test_list1) == 1
        assert test_list1[0].get() == "other test"
        callback.assert_not_called()

        # Test restoring a value that requires extending the List.

        test_list1.restoreField(".5", "extension test")
        assert len(test_list1) == 5
        assert test_list1[4].get() == "extension test"
        callback.assert_not_called()

        # Test different kinds of invalid restore paths.

        test_list2 = List(Key("default"))
        test_list2.onUpdateCall(callback)

        test_list2.restoreField("", "test")
        assert len(test_list2) == 0
        callback.assert_not_called()

        test_list2.restoreField("invalid", "test")
        assert len(test_list2) == 0
        callback.assert_not_called()

        test_list2.restoreField("invalid.1", "test")
        assert len(test_list2) == 0
        callback.assert_not_called()

        test_list2.restoreField("1.invalid", "test")
        assert len(test_list2) == 0
        callback.assert_not_called()

        test_list2.restoreField(".", "test")
        assert len(test_list2) == 0
        callback.assert_not_called()

        test_list2.restoreField(".0", "test")
        assert len(test_list2) == 0
        callback.assert_not_called()

        # Test indirect restore path.

        class TestBundle(Bundle):
            str_list = List(Key("default"))

        bundle = TestBundle()
        bundle.str_list.onUpdateCall(callback)
        bundle.str_list.restoreField("str_list.1", "test")
        assert len(bundle.str_list) == 1
        assert bundle.str_list[0].get() == "test"
        callback.assert_not_called()

        # Test an invalid restore value.

        test_list3 = List(Key("default"))
        test_list3.onUpdateCall(callback)
        test_list3.restoreField(".size", "invalid")
        assert len(test_list3) == 0
        callback.assert_not_called()

        # Test restoring a List with empty items.

        test_list4 = List(Key("default"))
        test_list4.onUpdateCall(callback)
        test_list4.restoreField(".1", None)
        test_list4.restoreField(".2", "")
        test_list4.restoreField(".3", None)
        assert len(test_list4) == 3
        assert not test_list4[0].isSet()
        assert test_list4[1].isSet() and test_list4[1].get() == ""
        assert not test_list4[2].isSet()
        callback.assert_not_called()

    def test_persistence(self):

        # A List does not keep a reference to its parent or children.

        bundle_list: List[Key[int]] = List(Key(default=0))
        level1: List[Key[int]] = List(Key(default=0))
        level1.setParent(bundle_list)
        level2: List[Key[int]] = List(Key(default=0))
        level2.setParent(level1)

        assert level1.parent() is not None
        assert len(list(level1.children())) == 1

        del bundle_list
        del level2
        assert level1.parent() is None
        assert len(list(level1.children())) == 0

    def test_update_callback_called_on_list_contents_changed(
        self, mocker: MockerFixture
    ):

        callback = mocker.stub()

        key_list: List[Key[int]] = List(Key(default=0))

        key_list.onUpdateCall(callback)

        key_list.appendOne()
        callback.assert_called_once_with(key_list)
        callback.reset_mock()

        key_list.append(Key(default=0))
        callback.assert_called_once_with(key_list)
        callback.reset_mock()

        key_list.insert(0, Key(default=0))
        callback.assert_called_once_with(key_list)
        callback.reset_mock()

        key_list[0] = Key(default=0)
        callback.assert_called_once_with(key_list)
        callback.reset_mock()

        key_list += [Key(default=0), Key(default=0), Key(default=0)]
        callback.assert_called_once_with(key_list)
        callback.reset_mock()

        key_list.extend([Key(default=0), Key(default=0), Key(default=0)])
        callback.assert_called_once_with(key_list)
        callback.reset_mock()

        key_list[2:4] = [
            Key(default=0),
            Key(default=0),
            Key(default=0),
        ]
        callback.assert_called_once_with(key_list)
        callback.reset_mock()

        key_list.pop()
        callback.assert_called_once_with(key_list)
        callback.reset_mock()

        key = key_list[0]
        key_list.remove(key)
        assert key not in key_list
        callback.assert_called_once_with(key_list)
        callback.reset_mock()

        del key_list[0]
        callback.assert_called_once_with(key_list)
        callback.reset_mock()

        assert len(key_list[1:3]) == 2
        del key_list[1:3]
        callback.assert_called_once_with(key_list)
        callback.reset_mock()

        assert len(key_list) > 1
        key_list.clear()
        callback.assert_called_once_with(key_list)
        callback.reset_mock()

    def test_update_callback_called_on_contained_item_update(
        self, mocker: MockerFixture
    ):

        callback = mocker.stub()

        bundle_list: List[ExampleBundle] = List(ExampleBundle())
        bundle_list.onUpdateCall(callback)

        bundle1 = ExampleBundle()
        bundle_list.append(bundle1)
        callback.reset_mock()
        bundle1.test.set("test")
        callback.assert_called_once_with(bundle1.test)

        bundle2 = ExampleBundle()
        bundle_list.insert(0, bundle2)
        callback.reset_mock()
        bundle2.test.set("test")
        callback.assert_called_once_with(bundle2.test)

        bundle3 = ExampleBundle()
        bundle_list[0] = bundle3
        callback.reset_mock()
        bundle3.test.set("test")
        callback.assert_called_once_with(bundle3.test)

        bundle4 = ExampleBundle()
        bundle5 = ExampleBundle()
        bundle_list[1:2] = [bundle4, bundle5]
        callback.reset_mock()
        bundle4.test.set("test")
        callback.assert_called_once_with(bundle4.test)
        callback.reset_mock()
        bundle5.test.set("test")
        callback.assert_called_once_with(bundle5.test)

        bundle6 = ExampleBundle()
        bundle7 = ExampleBundle()
        bundle_list += [bundle6, bundle7]
        callback.reset_mock()
        bundle6.test.set("test")
        callback.assert_called_once_with(bundle6.test)
        callback.reset_mock()
        bundle7.test.set("test")
        callback.assert_called_once_with(bundle7.test)

        bundle8 = ExampleBundle()
        bundle9 = ExampleBundle()
        bundle_list.extend([bundle8, bundle9])
        callback.reset_mock()
        bundle8.test.set("test")
        callback.assert_called_once_with(bundle8.test)
        callback.reset_mock()
        bundle9.test.set("test")
        callback.assert_called_once_with(bundle9.test)

    def test_update_callback_not_called_for_removed_items(
        self, mocker: MockerFixture
    ):

        bundle_list: List[ExampleBundle] = List(ExampleBundle())

        bundle1 = ExampleBundle()
        bundle2 = ExampleBundle()

        callback = mocker.stub()
        bundle_list.onUpdateCall(callback)

        bundle1.test.clear()
        bundle_list[:] = [bundle1]
        del bundle_list[0]

        callback.reset_mock()
        bundle1.test.set("test")
        callback.assert_not_called()

        bundle1.test.clear()
        bundle2.test.clear()
        bundle_list[:] = [bundle1, bundle2]
        del bundle_list[0:2]

        callback.reset_mock()
        bundle1.test.set("test")
        bundle2.test.set("test")
        callback.assert_not_called()

        bundle1.test.clear()
        bundle2.test.clear()
        bundle_list[:] = [bundle1, bundle2]
        bundle_list.clear()

        callback.reset_mock()
        bundle1.test.set("test")
        bundle2.test.set("test")
        callback.assert_not_called()

        bundle1.test.clear()
        bundle_list[:] = [bundle1]
        bundle_list.pop()

        callback.reset_mock()
        bundle1.test.set("test")
        callback.assert_not_called()

        bundle1.test.clear()
        bundle_list[:] = [bundle1]
        bundle_list.remove(bundle1)
        assert bundle1 not in bundle_list

        callback.reset_mock()
        bundle1.test.set("test")
        callback.assert_not_called()

    def test_label_set_on_contained_items(self):

        # Testing List.appendOne()

        key_list = List(Key(""))
        key = key_list.appendOne()
        assert key.fieldLabel() == "1"

        # Testing List.insertOne()

        other_key = key_list.insertOne(0)
        assert other_key.fieldLabel() == "1"
        assert key.fieldLabel() == "2"

        # Testing List.append()

        key = Key("")
        assert key.fieldLabel() == ""
        key_list.append(key)
        assert key.fieldLabel() == "3"

        # Testing List.extend()

        key = Key("")
        last_key = Key("")
        key_list.extend((key, last_key))
        assert key.fieldLabel() == "4"
        assert last_key.fieldLabel() == "5"

        # Testing List.insert()

        key = Key("")
        key_list.insert(3, key)
        assert key.fieldLabel() == "4"
        assert last_key.fieldLabel() == "6"

        # Testing List.__setitem__(index)

        key = Key("")
        key_list[2] = key
        assert key.fieldLabel() == "3"
        assert last_key.fieldLabel() == "6"

        # Testing List.__setitem__(slice)

        key = Key("")
        other_key = Key("")
        key_list[1:4] = [key, other_key]
        assert key.fieldLabel() == "2"
        assert other_key.fieldLabel() == "3"
        assert last_key.fieldLabel() == "5"

        # Testing List.__iadd__()

        key = Key("")
        other_key = Key("")
        key_list += [key, other_key]
        assert key.fieldLabel() == "6"
        assert other_key.fieldLabel() == "7"

        # Testing assignment order

        key_list[2], key_list[4] = key_list[4], key_list[2]
        assert key_list[2].fieldLabel() == "3"
        assert key_list[4].fieldLabel() == "5"

    def test_label_unset_on_removed_items(self):

        key_list = List(Key(""))
        for _ in range(15):
            key_list.appendOne()

        # Test List.__delitem__(index)

        key = key_list[3]
        assert key.fieldLabel() == "4"
        del key_list[3]
        assert key.fieldLabel() == ""

        # Test List.__delitem__(slice)

        keys = key_list[2:5]
        assert [key.fieldLabel() for key in keys] == ["3", "4", "5"]
        del key_list[2:5]
        assert [key.fieldLabel() for key in keys] == ["", "", ""]

        # Test List.__setitem__(index)

        key = key_list[5]
        assert key.fieldLabel() == "6"
        key_list[5] = Key("")
        assert key.fieldLabel() == ""

        # Test List.__setitem__(slice)

        keys = key_list[4:7]
        assert [key.fieldLabel() for key in keys] == ["5", "6", "7"]
        key_list[4:7] = [Key(""), Key("")]
        assert [key.fieldLabel() for key in keys] == ["", "", ""]

        # Test List.pop()

        last_key = key_list[9]
        assert last_key.fieldLabel() == "10"
        key_list.pop()
        assert last_key.fieldLabel() == ""

        # Test List.clear()

        keys = key_list[:]
        key_list.clear()
        assert all(key.fieldLabel() == "" for key in keys)

    def test_field_path(self):

        test_list1 = List(Key(""))
        assert test_list1.fieldPath() == "."
        test_list1.appendOne()
        test_list1.appendOne()
        assert test_list1[0].fieldPath() == ".1"
        assert test_list1[1].fieldPath() == ".2"

        test_list2 = List(ExampleBundle())
        assert test_list2.fieldPath() == "."
        test_list2.appendOne()
        test_list2.appendOne()
        assert test_list2[0].fieldPath() == ".1."
        assert test_list2[1].fieldPath() == ".2."
        assert test_list2[0].test.fieldPath() == ".1.test"
        assert test_list2[1].test.fieldPath() == ".2.test"

    def test_repr(self):

        parent = List(Key(default=0))
        list_iter_no_parent = List(Key(default=0), order=List.NO_PARENT)
        list_iter_parent_first = List(Key(default=0), order=List.PARENT_FIRST)
        list_iter_parent_last = List(Key(default=0), order=List.PARENT_LAST)

        list_iter_no_parent.setParent(parent)
        list_iter_parent_first.setParent(parent)
        list_iter_parent_last.setParent(parent)

        list_iter_no_parent.appendOne().set(1)
        list_iter_parent_first.appendOne().set(1)
        list_iter_parent_last.appendOne().set(1)

        assert repr(list_iter_no_parent) == "[<Key[int]:1>]"
        assert repr(list_iter_parent_first) == "[<parent>,<Key[int]:1>]"
        assert repr(list_iter_parent_last) == "[<Key[int]:1>,<parent>]"



from typing import Optional

import pytest

from pytest_mock import MockerFixture

from sunset import Bundle, Key, protocols


class ExampleSerializable:
    def __init__(self, value: str) -> None:

        self._value = value

    def toStr(self) -> str:

        return self._value

    @staticmethod
    def fromStr(value: str) -> Optional["ExampleSerializable"]:

        return ExampleSerializable(value)


class TestKey:
    def test_protocol_implementation(self):

        key = Key(default="")
        assert isinstance(key, protocols.Field)

    def test_default(self):

        assert Key(default="default").get() == "default"
        assert Key(default=0).get() == 0
        assert type(Key(default=False).get()) is bool
        assert not Key(default=False).get()
        assert Key(default=12.345e-67).get() == 12.345e-67
        with pytest.raises(TypeError):
            Key(default=object())  # type: ignore # It's the point!

    def test_set(self):

        key = Key(default="test")
        key.set("other")
        assert key.get() == "other"

    def test_clear(self):

        key = Key(default="default")
        key.set("other")
        assert key.get() != "default"

        key.clear()
        assert key.get() == "default"

    def test_serializable_type(self):

        value = ExampleSerializable.fromStr("dummy")
        assert value is not None
        key: Key[ExampleSerializable] = Key(default=value)

        assert key.get().toStr() == "dummy"

    def test_value_change_callback(self, mocker: MockerFixture):

        callback = mocker.stub()

        key = Key(default="default")

        key.onValueChangeCall(callback)

        key.set("default")
        callback.assert_not_called()
        callback.reset_mock()

        key.set("not default")
        callback.assert_called_once_with("not default")
        callback.reset_mock()

        key.clear()
        callback.assert_called_once_with("default")
        callback.reset_mock()

        key.clear()
        callback.assert_not_called()

    def test_value_change_callback_inheritance(self, mocker: MockerFixture):

        callback_sub_child1 = mocker.stub()
        callback_sub_child2 = mocker.stub()

        # Set up keys so that one parent key has two children key, each with one
        # child of its own.

        parent_key = Key(default="default")
        child1_key = Key(default="default")
        child2_key = Key(default="default")
        sub_child1_key = Key(default="default")
        sub_child2_key = Key(default="default")

        child1_key.setParent(parent_key)
        child2_key.setParent(parent_key)
        sub_child1_key.setParent(child1_key)
        sub_child2_key.setParent(child2_key)

        # Set up callbacks on the grandchildren keys.

        sub_child1_key.onValueChangeCall(callback_sub_child1)
        sub_child2_key.onValueChangeCall(callback_sub_child2)

        # One of the first generation children gets a value set on it. That
        # should count as set even if the value is the same as the default value
        # of the parent! But the sub-child's apparent value does not change, so
        # its callback shouldn't be called.

        child1_key.set("default")
        callback_sub_child1.assert_not_called()

        # If the toplevel parent's value changes to something new, the apparent
        # value change should be propagated down the inheritance chain through
        # all *unset* children. So only the second one in this case.

        parent_key.set("test")
        callback_sub_child1.assert_not_called()
        callback_sub_child2.assert_called_once_with("test")
        callback_sub_child2.reset_mock()

        # Updating the intermediate child should update the value of its own
        # child.

        child1_key.set("test")
        callback_sub_child1.assert_called_once_with("test")
        callback_sub_child1.reset_mock()

        # Clearing it causes its child to now inherit the value from the
        # toplevel parent, but that value is the same; so the child's value is
        # not updated.

        child1_key.clear()
        callback_sub_child1.assert_not_called()
        callback_sub_child1.reset_mock()

        # Clearing the parent does change the value inherited by the grandchild.

        parent_key.clear()
        callback_sub_child1.assert_called_once_with("default")
        callback_sub_child1.reset_mock()

    def test_key_updated_callback(self, mocker: MockerFixture):

        callback = mocker.stub()

        key = Key(default="default")
        assert isinstance(key, protocols.UpdateNotifier)

        key.onUpdateCall(callback)

        key.set("default")
        callback.assert_called_once_with(key)
        callback.reset_mock()

        key.set("not default")
        callback.assert_called_once_with(key)
        callback.reset_mock()

        key.set("not default")
        callback.assert_not_called()
        callback.reset_mock()

        key.clear()
        callback.assert_called_once_with(key)
        callback.reset_mock()

        key.clear()
        callback.assert_not_called()
        callback.reset_mock()

    def test_inheritance(self):

        parent_key = Key(default="default a")
        child_key = Key(default="")

        assert child_key not in parent_key.children()
        child_key.setParent(parent_key)
        assert child_key in parent_key.children()

        assert child_key.get() == "default a"

        parent_key.set("new a")
        assert child_key.get() == "new a"

        child_key.set("new b")
        assert child_key.get() == "new b"

        parent_key.clear()
        assert child_key.get() == "new b"

        child_key.clear()
        assert child_key.get() == "default a"

    def test_inherit_wrong_type(self):

        parent_key = Key(default="str")
        child_key = Key(default=0)

        # Ignore the type error, as it's the whole point of the test.

        child_key.setParent(parent_key)  # type: ignore

        assert child_key.parent() is None

    def test_inherit_revert(self):

        parent_key = Key(default="default a")
        child_key = Key(default="default b")

        assert child_key not in parent_key.children()
        assert child_key.parent() is None

        child_key.setParent(parent_key)

        assert child_key in parent_key.children()
        assert child_key.parent() is parent_key

        child_key.setParent(None)

        assert child_key not in parent_key.children()
        assert child_key.parent() is None

    def test_repr(self):

        key1 = Key(default="test")
        key2 = Key(default=12)
        key3 = Key(default="  test\ntest  ")

        assert repr(key1) == "<Key[str]:(test)>"
        assert repr(key2) == "<Key[int]:(12)>"
        assert repr(key3) == '<Key[str]:("  test\\ntest  ")>'

        key1.set("test")
        key2.set(12)
        key3.set("  test\ntest  ")

        assert repr(key1) == "<Key[str]:test>"
        assert repr(key2) == "<Key[int]:12>"
        assert repr(key3) == '<Key[str]:"  test\\ntest  ">'

    def test_reparenting(self):

        key1 = Key(default="default a")
        key2 = Key(default="default b")
        child_key = Key(default="default c")

        assert child_key not in key1.children()
        assert child_key not in key2.children()

        child_key.setParent(key1)
        assert child_key in key1.children()
        assert child_key not in key2.children()

        child_key.setParent(key2)
        assert child_key not in key1.children()
        assert child_key in key2.children()

        child_key.setParent(None)
        assert child_key not in key1.children()
        assert child_key not in key2.children()

    def test_callback_triggered_on_parent_value_change(
        self, mocker: MockerFixture
    ):
        stub = mocker.stub()

        parent_key = Key(default="default a")
        child_key = Key(default="default b")
        child_key.setParent(parent_key)

        child_key.onValueChangeCall(stub)

        child_key.set("test 1")
        stub.assert_called_once_with("test 1")
        stub.reset_mock()

        child_key.clear()
        stub.assert_called_once_with("default a")
        stub.reset_mock()

        parent_key.set("test 2")
        stub.assert_called_once_with("test 2")
        stub.reset_mock()

        parent_key.clear()
        stub.assert_called_once_with("default a")

    def test_field_label(self):
        class TestBundle(Bundle):
            key1 = Key("test")
            key2 = Key("test")

        bundle = TestBundle()
        assert bundle.key1.fieldLabel() == "key1"
        assert bundle.key2.fieldLabel() == "key2"

        key = Key("test")
        assert key.fieldLabel() == ""

    def test_field_path(self):

        assert Key("").fieldPath() == ""

        class TestBundle(Bundle):
            key1 = Key("test")
            key2 = Key("test")

        bundle = TestBundle()
        assert bundle.key1.fieldPath() == ".key1"
        assert bundle.key2.fieldPath() == ".key2"

    def test_dump_fields(self):

        # An unattached Key should get dumped. It just doesn't have a label.

        key = Key("")
        assert list(key.dumpFields()) == []
        key.set("test")
        assert list(key.dumpFields()) == [("", "test")]

        class TestBundle(Bundle):

            str_key = Key(default="")
            serializable_key = Key(default=ExampleSerializable("empty"))
            _private = Key(default=0)

        bundle = TestBundle()

        # A public Key should get dumped.

        assert list(bundle.str_key.dumpFields()) == []
        bundle.str_key.set("test")
        assert list(bundle.str_key.dumpFields()) == [(".str_key", "test")]

        # A Key's value should be serialized in its dump.

        assert list(bundle.serializable_key.dumpFields()) == []
        bundle.serializable_key.set(ExampleSerializable("not empty"))
        assert list(bundle.serializable_key.dumpFields()) == [
            (".serializable_key", "not empty")
        ]

        # A Key with a private label should not get dumped.

        assert list(bundle._private.dumpFields()) == []  # type: ignore
        bundle._private.set(111)  # type: ignore
        assert list(bundle._private.dumpFields()) == []  # type: ignore

    def test_restore_field(self, mocker: MockerFixture):

        key: Key[int] = Key(0)
        callback = mocker.stub()
        key.onUpdateCall(callback)

        # The value should be set if the given label matches that of the Key. In
        # this case, "". In all other cases the value should not be set. As
        # well, restoring a field should not trigger a callback.

        assert not key.isSet()
        key.restoreField("", "1")
        assert key.isSet()
        assert key.get() == 1
        callback.assert_not_called()

        key.clear()
        callback.reset_mock()
        key.restoreField("invalid", "1")
        assert not key.isSet()
        callback.assert_not_called()
        key.restoreField(".invalid", "1")
        assert not key.isSet()
        callback.assert_not_called()
        key.restoreField("invalid.", "1")
        assert not key.isSet()
        callback.assert_not_called()
        key.restoreField(".", "1")
        assert not key.isSet()
        callback.assert_not_called()

        # If multiple values are set, the last one sticks.

        key.clear()
        callback.reset_mock()
        key.restoreField("", "1")
        key.restoreField("", "2")
        assert key.get() == 2
        callback.assert_not_called()

        # Invalid values do not update the Key.

        key.clear()
        callback.reset_mock()
        key.restoreField("", "?")
        assert not key.isSet()
        callback.assert_not_called()
        key.restoreField("", "")
        assert not key.isSet()
        callback.assert_not_called()
        key.restoreField("", None)
        assert not key.isSet()
        callback.assert_not_called()

        class TestBundle(Bundle):
            str_key = Key("")

        # Same as the above when the Key has a non-empty label.

        bundle = TestBundle()
        bundle.str_key.restoreField("", "test")
        assert not bundle.str_key.isSet()
        bundle.str_key.restoreField("test", "test")
        assert not bundle.str_key.isSet()
        bundle.str_key.restoreField(".str_key", "test")
        assert not bundle.str_key.isSet()
        bundle.str_key.restoreField("str_key.", "test")
        assert not bundle.str_key.isSet()

        bundle.str_key.restoreField("str_key", "test")
        assert bundle.str_key.isSet()
        assert bundle.str_key.get() == "test"

    def test_restore_field_serialization(self):

        key_str: Key[str] = Key(default="")
        key_str.restoreField("", "test")
        assert key_str.get() == "test"

        key_int: Key[int] = Key(default=0)
        key_int.restoreField("", "12")
        assert key_int.get() == 12

        key_float: Key[float] = Key(default=1.2)
        key_float.restoreField("", "3.4")
        assert key_float.get() == 3.4

        key_bool: Key[bool] = Key(default=False)
        key_bool.restoreField("", "true")
        assert key_bool.get()

        key_custom: Key[ExampleSerializable] = Key(
            default=ExampleSerializable("")
        )
        key_custom.restoreField("", "test")
        assert key_custom.get().toStr() == "test"

    def test_persistence(self):

        # A Key does not keep a reference to its parent or children.

        key: Key[str] = Key(default="")
        level1: Key[str] = Key(default="")
        level1.setParent(key)
        level2: Key[str] = Key(default="")
        level2.setParent(level1)

        assert level1.parent() is not None
        assert len(list(level1.children())) == 1

        del key
        del level2
        assert level1.parent() is None
        assert len(list(level1.children())) == 0



import weakref

from typing import Any, Iterator, MutableMapping, MutableSet, Type, TypeVar

_T = TypeVar("_T")


class NonHashableSet(MutableSet[_T]):
    """
    An implementation of a set that can contain non-hashable elements.

    Elements that are not hashable are distinguished by their id.
    """

    _contents: MutableMapping[int, _T]

    def __init__(
        self, mapping_type: Type[MutableMapping[int, _T]] = dict
    ) -> None:

        super().__init__()

        self._contents = mapping_type()

    def _computeHash(self, value: Any) -> int:

        # Try to return the normal hash for the value if possible. This is
        # because hash is more selective than id. For instance! If value is a
        # weakref, then hash properly identifies two distinct weakrefs to the
        # same object as equivalent, whereas id does not.
        # Sadly testing whether the value implements the Hashable proto does not
        # suffice. Dataclasses, for instance, implement the Hashable proto but
        # are not necessarily hashable.

        try:
            return hash(value)
        except TypeError:
            return id(value)

    def add(self, value: _T) -> None:

        self._contents[self._computeHash(value)] = value

    def discard(self, value: _T) -> None:

        try:
            del self._contents[self._computeHash(value)]
        except KeyError:
            pass

    def __contains__(self, object: Any) -> bool:

        return self._computeHash(object) in self._contents

    def __iter__(self) -> Iterator[_T]:

        yield from self._contents.values()

    def __len__(self) -> int:

        return len(self._contents)


class WeakNonHashableSet(NonHashableSet[_T]):
    """
    An implementation of a weak set that can contain non-hashable elements.

    Elements that are not hashable are distinguished by their id.
    """

    def __init__(self) -> None:

        super().__init__(mapping_type=weakref.WeakValueDictionary)
